/* PROC2.C
 *
 * Sample program using the DosQPocStatus() function
 * for OS/2 2.0
 *
 * Kai Uwe Rommel - Wed 25-Mar-1992
 *
 * can be compiled with
 * - 16-bit compiler in small and large data models (MS C 6.00 tested)
 * - 32-bit compiler (IBM C Set/2 tested)
 */

#define INCL_NOPM
#define INCL_DOSPROCESS
#define INCL_DOSMODULEMGR
#define INCL_DOSMEMMGR
#include <os2.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#if defined(__32BIT__)
  #define PTR(ptr, ofs)  ((void *) ((char *) (ptr) + (ofs)))
#else
  #define APIENTRY16 APIENTRY
  #if defined(M_I86SM) || defined(M_I86MM)
    #define PTR(ptr, ofs)  ((void *) ((char *) (ptr) + (ofs)))
  #else
    #define PTR(ptr, ofs)  ((void *) ((char *) (((ULONG) procstat & 0xFFFF0000) | (USHORT) (ptr)) + (ofs)))
    /* kludge to transform 0:32 into 16:16 pointer in this special case */
  #endif
#endif



/* DosQProcStatus() = DOSCALLS.154 */
USHORT APIENTRY16 DosQProcStatus(PVOID pBuf, USHORT cbBuf);

/* DosGetPrty = DOSCALLS.9 */
USHORT APIENTRY16 DosGetPrty(USHORT usScope, PUSHORT pusPriority, USHORT pid);


struct procstat
{
  ULONG  summary;
  ULONG  processes;
  ULONG  semaphores;
  ULONG  unknown1;
  ULONG  sharedmemory;
  ULONG  modules;
  ULONG  unknown2;
  ULONG  unknown3;
};


struct process
{
  ULONG  type;
  ULONG  threadlist;
  USHORT processid;
  USHORT parentid;
  USHORT sessiontype;
  USHORT unknown1;
  ULONG  unknown2;
  USHORT sessionid;
  USHORT unknown3;
  USHORT modulehandle;
  USHORT threads;
  /* lots of other unknown data */
};


struct thread
{
  ULONG  unknown1;
  USHORT threadid;
  USHORT threadsysid;
  ULONG  blockid;
  USHORT priority;
  USHORT unknown2;
  ULONG  unknown3;
  ULONG  unknown4;
  USHORT status;
  USHORT unknown5;
};


struct module
{
  ULONG  nextmodule;
  USHORT modhandle;
  USHORT modtype;
  ULONG  submodules;
  ULONG  segments;
  ULONG  unknown1;
  ULONG  namepointer;
  USHORT submodule[1];  /* varying, see member submodules */
};


struct semaphore
{
  ULONG  nextsem;
  USHORT owner;
  UCHAR  flag;
  UCHAR  refs;
  UCHAR  requests;
  UCHAR  unknown1;
  USHORT unknown2;
  ULONG  unknown3;
  USHORT index;
  USHORT dummy;
  UCHAR  name[1];       /* varying */
};


struct shmem
{
  ULONG  nextseg;
  USHORT handle;
  USHORT selector;
  USHORT refs;
  UCHAR  name[1];       /* varying */
};


struct procstat *procstat;


void proctree(USHORT pid, USHORT indent)
{
  struct process *proc;
  UCHAR name[256];
  USHORT prty = 0;

  for ( proc = PTR(procstat -> processes, 0);
        proc -> type != 3; /* not sure if there isn't another termination */
        proc = PTR(proc -> threadlist,                          /* method */
                   proc -> threads * sizeof(struct thread))
      )
  {
    if ( proc -> parentid == pid && proc -> type != 0 )
    {
      DosQueryModuleName(proc -> modulehandle, sizeof(name), name);
      DosGetPrty(PRTYS_PROCESS, &prty, proc -> processid);

      printf("%5d  %5d   %02X  %4d %04X %*s%s\n",
        proc -> processid, proc -> parentid, proc -> sessionid,
        proc -> threads, prty, indent, "", name);

      proc -> type = 0;  /* kludge, mark it as already printed */
      proctree(proc -> processid, indent + 2);
    }
  }

  if ( pid != 0 )
    return;

  /* if at the root level, check for those processes that have lost *
   * their parent process and show tham as if they were childs of 0 */

  for ( proc = PTR(procstat -> processes, 0);
        proc -> type != 3; /* not sure if there isn't another termination */
        proc = PTR(proc -> threadlist,                          /* method */
                   proc -> threads * sizeof(struct thread))
      )
  {
    if ( proc -> type != 0 )
    {
      DosQueryModuleName(proc -> modulehandle, sizeof(name), name);
      DosGetPrty(PRTYS_PROCESS, &prty, proc -> processid);

      printf("%5d  %5d   %02X  %4d %04X %*s%s\n",
        proc -> processid, proc -> parentid, proc -> sessionid,
        proc -> threads, prty, indent, "", name);

      proctree(proc -> processid, indent + 2);
    }

    proc -> type = 1;    /* kludge, reset mark */
  }
}


void threadlist(void)
{
  struct process *proc;
  struct thread *thread;
  UCHAR name[256];
  USHORT count;

  for ( proc = PTR(procstat -> processes, 0);
        proc -> type != 3;
        proc = PTR(proc -> threadlist,
                   proc -> threads * sizeof(struct thread))
      )
  {
    DosQueryModuleName(proc -> modulehandle, sizeof(name), name);
    printf("%s\n", name);

    for ( count = 0, thread = PTR(proc -> threadlist, 0);
          count < proc -> threads; count++, thread++ )
      printf("%5d  %5d  %04X %08lX\n", thread -> threadid,
             thread -> threadsysid, thread -> priority, thread -> blockid);
  }
}


void modlist(int imports)
{
  struct module *mod;
  UCHAR name[256];
  ULONG count;

  for ( mod = PTR(procstat -> modules, 0); ;
        mod = PTR(mod -> nextmodule, 0)
      )
  {
    if ( !imports || mod -> submodules )
      if ( imports )
      {
        printf("%04X  %s\n", mod -> modhandle, PTR(mod -> namepointer, 0));

        for ( count = 0; count < mod -> submodules; count++ )
        {
          DosQueryModuleName(mod -> submodule[count], sizeof(name), name);
          printf("  %04X  %s\n", mod -> submodule[count], name);
        }
      }
      else
        printf(" %04X   %2d  %7ld  %6ld  %s\n", mod -> modhandle,
               mod -> modtype ? 32 : 16, mod -> segments,
               mod -> submodules, PTR(mod -> namepointer, 0));

    if ( mod -> nextmodule == 0L )
      break;
  }
}


void semlist(void)
{
  struct semaphore *sem;

  for ( sem = PTR(procstat -> semaphores, 16);
        sem -> nextsem != 0L;
        sem = PTR(sem -> nextsem, 0)
      )
    printf("%3d  %3d   %02X  %4d   %04X  \\SEM%s\n",
           sem -> refs, sem -> requests, sem -> flag,
           sem -> owner, sem -> index, sem -> name);
}


void shmlist(void)
{
  struct shmem *shmem;
  UCHAR name[256];
  ULONG size;
#if defined(__32BIT__)
  ULONG attrib;
  PVOID base;
#else
  SEL sel;
#endif

  for ( shmem = PTR(procstat -> sharedmemory, 0);
        shmem -> nextseg != 0L;
        shmem = PTR(shmem -> nextseg, 0)
      )
  {
    strcpy(name, "\\SHAREMEM\\");
    strcat(name, shmem -> name);

#if defined(__32BIT__)
    DosGetNamedSharedMem(&base, name, PAG_READ);
    size = -1;
    attrib = PAG_SHARED | PAG_READ;
    if ( DosQueryMem(base, &size, &attrib) )
      size = 0;
    DosFreeMem(base);
#else
    DosGetShrSeg(name, &sel);
    if ( DosSizeSeg(sel, &size) )
      size = 0;
    DosFreeSeg(sel);
#endif

    printf(" %04X   %04X   %08lX  %4d  %s\n", shmem -> handle,
           shmem -> selector, size, shmem -> refs, name);
  }
}


void main(void)
{
  procstat = malloc(0x8000);
  DosQProcStatus(procstat, 0x8000);

  printf("\n Process list\n");
  printf("\n PID    PPID  SESS THRD PRIO PROGRAM");
  printf("\n------ ------ ---- ---- ---- ------->\n");

  proctree(0, 0);

  printf("\n Thread list\n");
  printf("\n TID    TSID  PRIO BLOCKID");
  printf("\n------ ------ ---- --------\n");

  threadlist();

  printf("\n Module list\n");
  printf("\nHANDLE TYPE SEGMENTS IMPORTS PATH");
  printf("\n------ ---- -------- ------- ---->\n");

  modlist(0);

  printf("\n Module tree\n");
  printf("\n--------------\n");

  modlist(1);

  printf("\n Semaphore list\n");
  printf("\nREF  REQ  FLAG OWNER INDEX  NAME");
  printf("\n---- ---- ---- ----- ------ ---->\n");

  semlist();

  printf("\n Shared memory list\n");
  printf("\nHANDLE  SEL    SEGSIZE    REF  NAME");
  printf("\n------ ------ ---------- ----- ---->\n");

  shmlist();

  exit(0);
}


/* End of PROCS.C */
